#define MOZ_SOURCE_STAMP cc272f7d48d3544ffaf242b51430ffcf3a932a29
#define MOZ_SOURCE_REPO https://hg.mozilla.org/releases/mozilla-release
#define MOZ_SOURCE_URL https://hg.mozilla.org/releases/mozilla-release/rev/cc272f7d48d3544ffaf242b51430ffcf3a932a29
